/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2008 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/

#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/skbuff.h>
#include <asm/atomic.h>
#include <asm/mach_avm.h>

#if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
#include <linux/avm_led_event.h>
#endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/
#if defined(CONFIG_AVM_LED)
#include <linux/avm_led.h>
#endif /*--- #if defined(CONFIG_AVM_LED) ---*/

#include <linux/jiffies.h>
#if defined(CONFIG_AVM_POWER)
#include <linux/avm_power.h>
#endif /*--- #if defined(CONFIG_AVM_POWER) ---*/

#if !defined(CONFIG_NETCHIP_ADM69961)
#define CONFIG_NETCHIP_ADM69961
#endif

#include <asm/mips-boards/prom.h>
#include <linux/avm_event.h>
#include <linux/avm_cpmac.h>
#include <linux/adm_reg.h>
#include "cpmac_if.h"
#include "cpmac_const.h"
#include "cpmac_debug.h"
#include "cpmac_eth.h"
#include "cpphy_const.h"
#include "cpphy_types.h"
#include "cpphy_mdio.h"
#include "adm6996.h"
#include "tantos.h"
#include "cpphy_adm6996.h"
#include "cpphy_ar8216.h"
#include "cpphy_mgmt.h"
#include <linux/ar_reg.h>


static cpphy_mdio_t *power_mdio; /* FIXME Is there a cleaner solution? */

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_power_init(cpphy_mdio_t *mdio) {
    unsigned int port;

    power_mdio = mdio;
    /* Prepare a default timeout */
    mdio->adm_power.timeout = CPMAC_TICK_LENGTH;
    if(!(mdio->Mode & CPPHY_SWITCH_MODE_READ)) {
        /* The switch is not accessible (e.g. ProfiVoIP CPU 2) */
        return;
    }

    /* Initialize the power management for the PHYs */
    mdio->adm_power.setup.time_on  = CPMAC_ADM_PHY_PUP_TO;
    mdio->adm_power.setup.time_off = CPMAC_ADM_PHY_PDN_TO;
    for(port = 0; port < mdio->switch_ports; port++) {
        mdio->adm_power.setup.mode[port] = ADM_PHY_POWER_ON;
        mdio->adm_power.port_on[port] = 1;
    }
    adm_switch_port_power_config(mdio, &mdio->adm_power.setup, 1);
    /* There is no handle from the power management, therefor use it like this. */
#   if defined(CONFIG_AVM_POWER)
    mdio->power_handle = PowerManagmentRegister("ethernet", adm_power_config);
#   endif /*--- #if defined(CONFIG_AVM_POWER) ---*/
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int cpphy_mgmt_check_link(cpphy_mdio_t *mdio) {
    unsigned int device, port, old_linked = mdio->linked;
#   if defined(CONFIG_AVM_POWERMETER)
    unsigned int portset = 0, powervalue = 100;
#   endif /*--- #if defined(CONFIG_AVM_POWERMETER) ---*/

    switch(mdio->cpmac_switch) {
        case AVM_CPMAC_SWITCH_ADM6996:
        case AVM_CPMAC_SWITCH_ADM6996L:
        case AVM_CPMAC_SWITCH_TANTOS:
        case AVM_CPMAC_SWITCH_ANY:
            /* There is no checking of the link without read access */
            adm_check_link(mdio);
            break;

        case AVM_CPMAC_SWITCH_AR8216:
            ar_check_link(mdio);
            break;

        case AVM_CPMAC_SWITCH_NONE:
        default:
            return mdio->linked;
    }

    if(old_linked != mdio->linked) {
        cpphy_mgmt_event_dataupdate(mdio); /* Linked ports could have changed */
    }

    for(port = 0; port < TANTOS_MAX_PORTS ; port++) {
        unsigned int port_linked = (mdio->linked & (1 << port)) ? 1 : 0;
        if(port < CPMAC_MAX_LED_HANDLES) {
            if(port_linked != mdio->cpmac_priv->led[port].on) {
                mdio->cpmac_priv->led[port].on = port_linked;
                DEB_INFOTRC("Set LAN LED %u to '%s'\n",
                            port,
                            port_linked ? "on" : "off");
#               if defined(CONFIG_AVM_LED)
                if(mdio->cpmac_priv->led[port].handle != 0) {
                    avm_led_action_with_handle(mdio->cpmac_priv->led[port].handle,
                                               port_linked ? avm_led_on : avm_led_off);
                }
#               endif /*--- #if defined(CONFIG_AVM_LED) ---*/
#               if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
                if(led_event_action) {
                    (*led_event_action)(LEDCTL_VERSION, event_lan1_active + (port << 1) + (port_linked ? 0 : 1), 1);
                }
#               endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/
            }
        }
#       if defined(CONFIG_AVM_POWERMETER)
        if((port < mdio->switch_ports) && (mdio->adm_power.port_on)) {
            powervalue += 10; /* Port is powered */
            if(port_linked) {
                /* Port is linked */
                portset |= (1 << port);
                powervalue += mdio->event_data.port[port].speed100 ? 57 : 90; /* 100 MBit/s? */
            }
        }
#       endif /*--- #if defined(CONFIG_AVM_POWERMETER) ---*/
    }
    cpphy_mgmt_power(mdio);
#   if defined(CONFIG_AVM_POWERMETER)
    PowerManagmentRessourceInfo(powerdevice_ethernet, PM_ETHERNET_PARAM(portset, powervalue));
#   endif /*--- #if defined(CONFIG_AVM_POWERMETER) ---*/

    for(device = 0; device < mdio->cpmac_priv->devices; device++) {
        unsigned int carrier = netif_carrier_ok(mdio->cpmac_priv->devs[device]);

        /*--- DEB_TEST("[cpphy_mgmt_check_link] '%s' carrier %s linked = %#x   %#x = ports\n", ---*/
                 /*--- mdio->cpmac_priv->devs[device]->name, ---*/ 
                 /*--- carrier ? "on " : "off", ---*/
                 /*--- mdio->linked, mdio->cpmac_priv->dev_ports[device]); ---*/
        if(mdio->linked & mdio->cpmac_priv->dev_ports[device]) {
            if(!carrier) {
                netif_carrier_on(mdio->cpmac_priv->devs[device]);
                DEB_INFOTRC("Switching carrier on for device %s\n", mdio->cpmac_priv->devs[device]->name);
            }
        } else if(carrier) {
            netif_carrier_off(mdio->cpmac_priv->devs[device]);
            DEB_INFOTRC("Switching carrier off for device %s\n", mdio->cpmac_priv->devs[device]->name);
        }
    }

    return mdio->linked;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_event_dataupdate(cpphy_mdio_t *mdio) {
    unsigned int port, changes = 0;
    /* Update LAN-LED-state */
    /* TODO Does this exist twice now? (Another time in check_link) */
    /* Switch LED control happens in cpphy_adm6996.c */
    if(!mdio->cpmac_switch) {
#       if defined(CONFIG_AVM_LED)
        if(mdio->cpmac_priv->led[mdio->cpmac_priv->inst].handle != 0) {
            avm_led_action_with_handle(mdio->cpmac_priv->led[mdio->cpmac_priv->inst].handle,
                                       (mdio->state == CPPHY_MDIO_ST_LINKED) ? avm_led_on
                                                                             : avm_led_off);
        }
#       endif /*--- #if defined(CONFIG_AVM_LED) ---*/
#       if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE)
        if(led_event_action) {
            (*led_event_action)(LEDCTL_VERSION,   event_lan1_active 
                                                + (mdio->cpmac_priv->inst << 1) 
                                                + (mdio->state == CPPHY_MDIO_ST_LINKED) ? 0 : 1, 1);
        }
#       endif /*--- #if defined(CONFIG_AVM_LED_EVENTS) || defined(CONFIG_AVM_LED_EVENTS_MODULE) ---*/
    }
    switch(mdio->cpmac_switch) {
        case AVM_CPMAC_SWITCH_ADM6996:
        case AVM_CPMAC_SWITCH_ADM6996L:
            break; /* At the moment done directly in adm_check_link */
            /* The data in the adm_serial_register got updated by a previous call to *\
            \* ADM_GET_SERIAL_REG(REG_ADM_LC_STATUS0)                                */
            /*--- changes |= mdio->event_data.port[0].link != adm_serial_register.status.port0linkup; ---*/
            /*--- changes |= mdio->event_data.port[1].link != adm_serial_register.status.port1linkup; ---*/
            /*--- changes |= mdio->event_data.port[2].link != adm_serial_register.status.port2linkup; ---*/
            /*--- changes |= mdio->event_data.port[3].link != adm_serial_register.status.port3linkup; ---*/
            /*--- changes |= mdio->event_data.port[4].link != adm_serial_register.status.port4linkup; ---*/
            /*--- if(changes) { ---*/
                /*--- mdio->event_data.port[0].link = adm_serial_register.status.port0linkup; ---*/
                /*--- mdio->event_data.port[0].speed100 = adm_serial_register.status.port0speed; ---*/
                /*--- mdio->event_data.port[0].fullduplex = adm_serial_register.status.port0duplex; ---*/
                /*--- mdio->event_data.port[1].link = adm_serial_register.status.port1linkup; ---*/
                /*--- mdio->event_data.port[1].speed100 = adm_serial_register.status.port1speed; ---*/
                /*--- mdio->event_data.port[1].fullduplex = adm_serial_register.status.port1duplex; ---*/
                /*--- mdio->event_data.port[2].link = adm_serial_register.status.port2linkup; ---*/
                /*--- mdio->event_data.port[2].speed100 = adm_serial_register.status.port2speed; ---*/
                /*--- mdio->event_data.port[2].fullduplex = adm_serial_register.status.port2duplex; ---*/
                /*--- mdio->event_data.port[3].link = adm_serial_register.status.port3linkup; ---*/
                /*--- mdio->event_data.port[3].speed100 = adm_serial_register.status.port3speed; ---*/
                /*--- mdio->event_data.port[3].fullduplex = adm_serial_register.status.port3duplex; ---*/
                /*--- mdio->event_data.port[4].link = adm_serial_register.status.port4linkup; ---*/
                /*--- mdio->event_data.port[4].speed100 = adm_serial_register.status.port4speed; ---*/
                /*--- mdio->event_data.port[4].fullduplex = adm_serial_register.status.port4duplex; ---*/
            /*--- } ---*/
            break;
        case AVM_CPMAC_SWITCH_TANTOS:
            for(port = 0; port < 5; port++) {
                changes |= mdio->event_data.port[port].link != tantos_ports->port[port].PS.link;
            }
            if(changes) {
                for(port = 0; port < 5; port++) {
                    mdio->event_data.port[port].link = tantos_ports->port[port].PS.link;
                    mdio->event_data.port[port].speed100 = tantos_ports->port[port].PS.speed;
                    mdio->event_data.port[port].fullduplex = tantos_ports->port[port].PS.duplex;
                }
            }
            break;
        case AVM_CPMAC_SWITCH_AR8216:
            for(port = 0; port < 4; port++) {
                unsigned int link, status;
                status = cpphy_mdio_user_access_read(mdio, AR8216_MII_PHY_STATUS, port);
                link = (status & ATHR_PHY_STATUS_LINK) ? 1 : 0;
                changes |= (link != mdio->event_data.port[port].link);
                mdio->event_data.port[port].link = link;
                mdio->event_data.port[port].speed100 = (((status & ATHR_PHY_STATUS_SPEED_MASK) >> ATHR_PHY_STATUS_SPEED_SHIFT) == 1) ? 1 : 0;
                mdio->event_data.port[port].fullduplex = (status & ATHR_PHY_STATUS_DUPLEX) ? 1 : 0;
            }
            break;
        case AVM_CPMAC_SWITCH_NONE:
            if(mdio->event_data.port[mdio->inst].link != (mdio->state == CPPHY_MDIO_ST_LINKED)) {
                /* Nothing to do for low phy when switch is connected to high cpmac */
                /* -> dont register low cpphy driver at cpmac */
                changes = 1;
                mdio->event_data.port[mdio->inst].link = (mdio->state == CPPHY_MDIO_ST_LINKED);
                mdio->event_data.port[mdio->inst].speed100 = !mdio->slow_speed;
                mdio->event_data.port[mdio->inst].fullduplex = !mdio->half_duplex;
            }
    }
    if(changes) {
        switch(mdio->cpmac_switch) {
            case AVM_CPMAC_SWITCH_ADM6996:
            case AVM_CPMAC_SWITCH_ADM6996L:
            case AVM_CPMAC_SWITCH_TANTOS:
            case AVM_CPMAC_SWITCH_AR8216:
                DEB_TRC("[cpphy_mgmt_event_dataupdate] Port stati: 0 %s%s%s 1 %s%s%s 2 %s%s%s 3 %s%s%s\n",
                        mdio->event_data.port[0].link ? "up" : "down",
                        mdio->event_data.port[0].link ? (mdio->event_data.port[0].speed100 ? " 100" : " 10") : "",
                        mdio->event_data.port[0].link ? (mdio->event_data.port[0].fullduplex ? " FD" : " HD") : "",
                        mdio->event_data.port[1].link ? "up" : "down",
                        mdio->event_data.port[1].link ? (mdio->event_data.port[1].speed100 ? " 100" : " 10") : "",
                        mdio->event_data.port[1].link ? (mdio->event_data.port[1].fullduplex ? " FD" : " HD") : "",
                        mdio->event_data.port[2].link ? "up" : "down",
                        mdio->event_data.port[2].link ? (mdio->event_data.port[2].speed100 ? " 100" : " 10") : "",
                        mdio->event_data.port[2].link ? (mdio->event_data.port[2].fullduplex ? " FD" : " HD") : "",
                        mdio->event_data.port[3].link ? "up" : "down",
                        mdio->event_data.port[3].link ? (mdio->event_data.port[3].speed100 ? " 100" : " 10") : "",
                        mdio->event_data.port[3].link ? (mdio->event_data.port[3].fullduplex ? " FD" : " HD") : "");
                break;
            case AVM_CPMAC_SWITCH_NONE:
            default:
                DEB_TRC("[cpphy_mgmt_event_dataupdate] Phy %u status: %s%s%s\n",
                        mdio->inst,
                        mdio->event_data.port[mdio->inst].link ? "up" : "down",
                        mdio->event_data.port[mdio->inst].link ? (mdio->event_data.port[mdio->inst].speed100 ? " 100" : " 10") : "",
                        mdio->event_data.port[mdio->inst].link ? (mdio->event_data.port[mdio->inst].fullduplex ? " FD" : " HD") : "");
                break;
        }
        cpmac_main_event_update();
    }
}


/*------------------------------------------------------------------------------------------*\
 * Switch port power, assume link status is uptodate
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_power(cpphy_mdio_t *mdio) {
    unsigned int port, from = 0, to = mdio->switch_ports - 1;
    unsigned int portset = 0;

    /* There is no power saving without write access */
    if(!(mdio->Mode & CPPHY_SWITCH_MODE_WRITE)) {
        mdio->adm_power.timeout = CPMAC_TICK_LENGTH;
        return;
    }

    /* Switch to different state than last run and use correct timeout*/
    mdio->adm_power.current_on = !mdio->adm_power.current_on;
    /* If global powerdown is needed, disable roundrobin for this run */
    if(mdio->global_power) {
        if(mdio->global_power == CPPHY_POWER_GLOBAL_ON) {
            DEB_INFOTRC("Global powerup active\n");
            mdio->adm_power.current_on = 1;
        } else if(mdio->global_power == CPPHY_POWER_GLOBAL_OFF) {
            DEB_INFOTRC("Global powerdown active\n");
            mdio->adm_power.current_on = 0;
        } else {
            DEB_ERR("[cpphy_mgmt_power] Internal error: Global power with unknown value %u\n",
                    mdio->global_power);
        }
    } else if(mdio->adm_power.roundrobin) {
        /* If we are going to activate a port, it should be the next one */
        if(mdio->adm_power.current_on) {
            mdio->adm_power.current_port = (mdio->adm_power.current_port == 4)
                                           ? 0
                                           : (mdio->adm_power.current_port + 1);
        }
        from = to = mdio->adm_power.current_port;
    }

    mdio->adm_power.timeout =   (mdio->global_power == CPPHY_POWER_GLOBAL_OFF)
                              ? CPMAC_TICK_LENGTH
                              : (mdio->adm_power.current_on) ? mdio->adm_power.setup.time_on
                                                             : mdio->adm_power.setup.time_off;

    for(port = from; port <= to; port++) {
        if(   !mdio->adm_power.port_on[port]     /* If the port is disabled,    */
           || !mdio->event_data.port[port].link  /* without link                */
           || mdio->global_power) {              /* or a global power requested */
            unsigned int turn_on = 0;
            if(   ((mdio->adm_power.setup.mode[port] == ADM_PHY_POWER_SAVE) && mdio->adm_power.current_on)
               || (mdio->adm_power.setup.mode[port] == ADM_PHY_POWER_ON)) {
                turn_on = 1;
            } else { /*--- (mdio->adm_power.setup.mode[port] == ADM_PHY_POWER_OFF) ---*/
                turn_on = 0;
            }
            switch(mdio->global_power) {
                case CPPHY_POWER_GLOBAL_ON:
                    if(   (mdio->adm_power.setup.mode[port] == ADM_PHY_POWER_SAVE)
                       || (mdio->adm_power.setup.mode[port] == ADM_PHY_POWER_ON)) {
                        turn_on = 1;
                    }
                    break;
                case CPPHY_POWER_GLOBAL_OFF:
                    turn_on = 0;
                    break;
                default:
                    break;
            }
            if(mdio->adm_power.port_on[port] != turn_on) {
                DEB_TEST("Power %s PHY %u\n", turn_on ? "up" : "down", port);
                switch(mdio->cpmac_switch) {
                    case AVM_CPMAC_SWITCH_ADM6996:
                    case AVM_CPMAC_SWITCH_ADM6996L:
                    case AVM_CPMAC_SWITCH_TANTOS:
                        adm_switch_port_power(mdio, port, turn_on);
                        break;
                    case AVM_CPMAC_SWITCH_AR8216:
                        if(mdio->global_power == CPPHY_POWER_GLOBAL_ON) {
                            portset |= 1 << port;
                            break;
                        }
                        ar_switch_port_power(mdio, port, turn_on);
                        break;
                    case AVM_CPMAC_SWITCH_ANY: /* How to do power control should be known */
                    case AVM_CPMAC_SWITCH_NONE:
                    default:
                        break;
                }
                mdio->adm_power.port_on[port] = turn_on;
            }
        }
    }
    if(mdio->global_power == CPPHY_POWER_GLOBAL_ON) {
        mdio->global_power = CPPHY_POWER_GLOBAL_NONE;
        if(mdio->cpmac_switch == AVM_CPMAC_SWITCH_AR8216) {
            ar_switch_port_power_on_boot(mdio, portset);
        }
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t adm_switch_port_power_config(cpphy_mdio_t          *mdio,
                                         adm_phy_power_setup_t *setup,
                                         unsigned char          set) {
    unsigned int i;

    if(!set) {
        setup->time_on  = mdio->adm_power.setup.time_on;
        setup->time_off = mdio->adm_power.roundrobin ? (4 * (  mdio->adm_power.setup.time_on
                                                             + mdio->adm_power.setup.time_off))
                                                     : mdio->adm_power.setup.time_off;
        for(i = 0; i < 5; i++) {
            setup->mode[i] = mdio->adm_power.setup.mode[i];
        }
    } else {

        mdio->adm_power.timeout = 0;
        mdio->adm_power.current_on = 1;
        mdio->adm_power.current_port = 0;

        mdio->adm_power.roundrobin = ((setup->time_on * 4) < setup->time_off);
        mdio->adm_power.setup.time_on  = setup->time_on;
        mdio->adm_power.setup.time_off =   mdio->adm_power.roundrobin
                                         ? ((setup->time_off - (4 * setup->time_on)) / 4)
                                         : setup->time_off;
        DEB_INFOTRC("Power setup: %s with %u on and %u off\n", 
                    mdio->adm_power.roundrobin ? "roundrobin" : "normal",
                    mdio->adm_power.setup.time_on,
                    mdio->adm_power.setup.time_off);
        for(i = 0; i < 5; i++) {
            mdio->adm_power.setup.mode[i] = setup->mode[i];
        }
    }
    return CPMAC_ERR_NOERR;
}


#if defined(CONFIG_AVM_POWER)
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int adm_power_config(int state) {
    union _powermanagment_ethernet_state pstate;

    pstate.Register = state;
    if((pstate.Bits.port < 0) || (pstate.Bits.port > 4)) {
        DEB_ERR("adm_power_config tried to set mode for port %u\n", pstate.Bits.port);
        return -1;
    }
    power_mdio->adm_power.setup.mode[pstate.Bits.port] = pstate.Bits.status;
    DEB_TRC("adm_power_config sets mode %u for port %u\n", pstate.Bits.status, pstate.Bits.port);
    return 0;
}
#endif /*--- #if defined(CONFIG_AVM_POWER) ---*/


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmac_work(void *data) {
    cpphy_mdio_t *mdio = (cpphy_mdio_t *) data;
    unsigned long next_run;
    unsigned long next_wait;
    unsigned int something_done, i;

    assert(mdio != NULL);
    for( ; ; ) {
        mdio->work_to_do = 0;
        if(mdio->stop_work) {
            DEB_TRC("[cpmac_work] Stop work\n");
            break;
        }
        something_done = 0;
        next_run = jiffies + 100000; /* Waiting should be a lot shorter than this */
        for(i = 0; i < CPMAC_WORK_MAX_ITEMS; i++) {
            if(mdio->work[i].active) {
                if(time_after_eq(jiffies, mdio->work[i].next_run)) { 
                    /* It is time to run this item */
                    assert(mdio->work[i].item);
                    mdio->work[i].next_run = jiffies + mdio->work[i].item(mdio);
                    something_done = 1;
                    break; /* Restart queue checking */
                }
                if(time_after(next_run, mdio->work[i].next_run)) {
                    next_run = mdio->work[i].next_run;
                }
            }
        }
        if(!something_done) {
            next_wait = next_run - jiffies;
            if(time_after(next_run, jiffies)) {
                wait_event_interruptible_timeout(mdio->waitqueue, mdio->work_to_do, next_wait);
            } else {
                DEB_TEST("[cpmac_work] Not waiting\n");
            }
        }
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_work_start(cpphy_mdio_t *mdio) {
    mdio->stop_work = 0;
    INIT_WORK(&mdio->cpmac_work, cpmac_work, (void *) mdio);
    cancel_delayed_work(&mdio->cpmac_work);
    queue_delayed_work(cpmac_global.workqueue, &mdio->cpmac_work, 0);
}


/*------------------------------------------------------------------------------------------*\
 * Add a function to be managed
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpphy_mgmt_work_add(cpphy_mdio_t        *mdio, 
                                cpmac_work_ident_t   ident,
                                cpmac_work_function  function,
                                unsigned long        initial_timeout) {
    DEB_INFOTRC("[cpphy_mgmt_work_add] Function for ID %u with timeout %lu added!\n", 
                ident,
                initial_timeout);
    if((mdio->work[ident].item != NULL) && (mdio->work[ident].item != function)) {
        DEB_ERR("[cpphy_mgmt_work_add] Different function for ID %u already added!\n", ident);
        return CPMAC_ERR_INTERNAL;
    }
    mdio->work[ident].item = function;
    if(time_after(mdio->work[ident].next_run, jiffies + initial_timeout)) {
        mdio->work[ident].next_run = jiffies + initial_timeout;
    }
    mdio->work[ident].active = 1;
    mdio->work_to_do = 1;
    wake_up_interruptible_sync(&mdio->waitqueue);
    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
 * Remove a function to be managed
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpphy_mgmt_work_del(cpphy_mdio_t       *mdio, 
                                cpmac_work_ident_t  ident) {
    if(mdio->work[ident].active == 0) {
        DEB_WARN("[cpphy_mgmt_work_del] Function for ID %u is not set!\n", ident);
        return CPMAC_ERR_INTERNAL;
    }
    mdio->work[ident].active = 0;
    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
 * Schedule the work item after the given timeout; 0 means as soon as possible
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_work_schedule(cpphy_mdio_t *mdio, 
                              cpmac_work_ident_t ident, 
                              unsigned long timeout) {
    mdio->work[ident].next_run = jiffies + timeout;
    mdio->work[ident].active = 1;
    mdio->work_to_do = 1;
    wake_up_interruptible_sync(&mdio->waitqueue);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_work_stop(cpphy_mdio_t *mdio) {
    mdio->stop_work = 1;
    mdio->work_to_do = 1;
    wake_up(&mdio->waitqueue);
    flush_workqueue(cpmac_global.workqueue);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
cpmac_err_t cpphy_mgmt_init(cpphy_mdio_t *mdio) {
    unsigned int i;

    init_waitqueue_head(&mdio->waitqueue);
    mdio->work_to_do = 1;
    for(i = 0; i < CPMAC_MAX_WORK_ITEMS; i++) {
        mdio->work[i].active = 0;
        mdio->work[i].item = NULL;
        mdio->work[i].next_run = 0;
    }
    return CPMAC_ERR_NOERR;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpphy_mgmt_deinit(cpphy_mdio_t *mdio) {
    cpphy_mgmt_work_stop(mdio);
}

